// (c)1995,1996 Jaimi R.R. McEntire. All Rights Reserved
//
// Portions of the Wadloading code Converted to Pascal
// from C++ source (c)1994 Jake Hill.

unit dxint;

interface

// limitations of the 4DX engine. Do not change these. To change limits,
// recompile the 4DX DLL, and then change these.
const dx_MaxSectors = 1024;
      dx_MaxActors  = 1024;
      dx_MaxActorFrames  = 32;
      dx_MaxActorDirs    = 16;
      dx_MaxEvents       = 32;
      dx_MaxActorType    = 32;
      dx_MaxVertexes     = 4096;
      dx_MaxDist         = 4096;
      dx_MaxFlats        = 64;
      dx_MaxBitmaps      = 1024;
      dx_MaxTextures     = 1024;
      dx_MaxPatches      = 64;

const dx_Angle360 = 1024;
      dx_Angle270 = 768;
      dx_Angle180 = 512;
      dx_Angle90  = 256;
      dx_Angle45  = 128;
      dx_Angle22h = 64;
      dx_Angle11q = 32;

// Data types for WAD files (from Doom/Heretic/Hexen etc.)
Type pWAD_Header = ^WAD_Header;
     WAD_Header = record
       signature: Array[0..3] of Char;
       num_entries: LongInt;
       foffset: LongInt;
     end {WAD_Header};

Type pDirectory_Entry = ^Directory_Entry;
     Directory_Entry = record
       foffset: LongInt;
       size: LongInt;
       name: Array[0..7] of Char;
     end {Directory_Entry};

Type pv_vertex = ^v_vertex;
     v_vertex = record
       x: SmallInt;
       y: SmallInt;
     end {v_vertex};

Type pv_line = ^v_line;
     v_line = record
       from, tovert: SmallInt;
       flags, special, tag: SmallInt;
       side: Array[0..1] of SmallInt;
     end {v_line};

Type pv_thing = ^v_thing;
     v_thing = record
       x, y: SmallInt;
       angle: Word;
       thingtype: SmallInt;
       attributes: SmallInt;
     end {v_thing};

Type pv_side = ^v_side;
     v_side = record
       tm_xoffset, tm_yoffset: SmallInt;
       upper_tx: Array[0..7] of Char;
       lower_tx: Array[0..7] of Char;
       main_tx: Array[0..7] of Char;
       sector: SmallInt;
    end {v_side};

Type pv_sector = ^v_sector;
     v_sector = record
       floor_ht: SmallInt;
       ceiling_ht: SmallInt;
       floor_tx: Array[0..7] of Char;
       ceiling_tx: Array[0..7] of Char;
       light: SmallInt;
       sectortype: SmallInt;
       trigger: SmallInt;
     end {v_sector};

// Data types for 4DX.
type
   pVertex = ^Vertex;
   Vertex = record
     x,y,n_x,n_y,r_x,r_y,startx,starty : single;
   end;

type SmallVert = record
     num : smallint;
     x,y : single;
   end;

type
   pLineDef = ^LineDef;
   LineDef = record
   fromvert,tovert : integer;
   x1,y1,x2,y2     : single;
   dx,dy           : single;
   rx1,ry1,rx2,ry2 : single;
   COS,sine        : single;      // which angle this is facing so we know when to draw floor etc.
   Len,dist        : single;
   temp1,temp2     : single;
   temp3,temp4     : single;

   dotype, dotag   : integer;
   sector, facing  : integer;
   texture         : integer;
   flags           : integer;
   lineno          : integer;
   TripFlags       : integer;

   trip1,trip2,trip3,trip4 : integer;
   Texture2, topTexture, topTexture2  : integer;
   ltype    : integer;
   BegCol   : integer;
   EndCol   : integer;
   nextline : pLineDef;
   unused   : integer;
end;

type pActor = ^Actor;
     Actor  = record
       flags : integer;
       x,y,height,eyeheight,physht,adjust,halfwit : single;
       angle,actortype,sector,currframe : integer;
       vertforce,velocity,range : single;
       damage, dambase,maxhp,curhp,used : integer;
       perc_attack,mood,target : integer;
       tag,dotype,dotag : integer;
       event, talkevent,latmov,frames : integer;
       number,mapnumber : integer;
       l : pLineDef;
     end;

type pSector = ^sector;
     Sector  = record
        Number          :integer;
        FloorTX,CeilTX  : integer;
        UnderTX         : integer;
        Floorht,ceilht,underht,HTOffset : single;
        flags           : integer;
        XaFloor,YaFloor : integer;
        XaCeil,YaCeil   : integer;
        tx,ty,bx,by     : integer;
        lightlevel      : integer;
        tag             : integer;
        tripflags       : integer;
        swing           : integer;
        firstl,lastl    : pLineDef;
        unused          : integer;
        end;

type ppatch = ^patch;
     patch = record
        x,y : integer;
        w,h : integer;
        buffer : pointer;
        bitmapno : integer;
     end;

type pMetaTexture = ^MetaTexture;
        MetaTexture = record
        h,w,numpatches,hmask,wmask : integer;
        number : integer;
        hf,wf  : single;
        patcharray : array [1..64] of ppatch;
     end;

type SmallText = record
        w,h,number,numpatches : integer;
        name : array [0..7] of char;
     end;

type SmallPatch = record
        x,y,bmno,txno : integer;
     end;

type SmallType = record
         num,frames,dirs : integer;
         name : array [0..7] of char;
     end;

type pbminfo = ^bminfo;
     bminfo = record
         buffer : pointer;
         w,h    : integer;
         wf,hf  : single;
         wmask,hmask : integer;
         meta   : integer;
     end;

// Interface to the 4dx DLL.
function  dx_init(w,h,pitch,orientation : integer;databuff,palette : pointer) : integer; stdcall; external '4dxdll.dll';
function  dx_autoload(datafile : pchar) : integer;  stdcall; external '4dxdll.dll'; // not recommended for normal users.
function  dx_render(mode : integer; shear : integer) : integer; stdcall; external '4dxdll.dll';
procedure dx_mapzoomlevel(zoomlevel : integer); stdcall; external '4dxdll.dll';
procedure dx_mapcolors(boundcol,normalcol,arrowcol : integer);  stdcall; external '4dxdll.dll';
procedure dx_deletesector(number : integer);  stdcall; external '4dxdll.dll';
procedure dx_buildtexture(pm:pMetaTexture;data:pointer;maxsize,flip : integer);  stdcall; external '4dxdll.dll';

function  dx_getflatname(number : integer) : pchar; stdcall; external '4dxdll.dll';
function  dx_getbitmapname(number : integer) : pchar; stdcall; external '4dxdll.dll';
function  dx_gettexturename(number : integer) : pchar; stdcall; external '4dxdll.dll';
function  dx_gettypename(number : integer) : pchar;  stdcall; external '4dxdll.dll';
function  dx_getbitmapinfo(number : integer) : pbminfo; stdcall; external '4dxdll.dll';
function  dx_freebitmap(number : integer) : integer; stdcall; external '4dxdll.dll';

function  dx_addbitmap(number,w,h,pitch,bytes : integer; data : pointer; bmname : pchar) : integer; stdcall; external '4dxdll.dll';
function  dx_addpatch (texture,bitmap,xoffset,yoffset : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_createtexture(number,w,h : integer; txname : pchar) : integer; stdcall; external '4dxdll.dll';
procedure dx_deletetexture(texture : integer); stdcall; external '4dxdll.dll';
function  dx_addtypeinfo(actortype,frames,dirs : integer; aname : pchar) : integer; stdcall; external '4dxdll.dll';
function  dx_addtypedata(actortype,frame,dir,texture,reverse : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_gettypedata(actortype,frame,dir : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_gettypedirs(actortype : integer) : integer;  stdcall; external '4dxdll.dll';
function  dx_gettypeframes(actortype : integer) : integer;  stdcall; external '4dxdll.dll';
function  dx_getclosestcolor(r,g,b : integer) : integer;   stdcall; external '4dxdll.dll';

function  dx_createflat(number,w,pitch: integer;address : pointer; flatname : pchar) : integer; stdcall; external '4dxdll.dll';
procedure dx_modifypatch(texture,patch,xoff,yoff : integer);  stdcall; external '4dxdll.dll';
function  dx_getmetatexture(texture : integer) : pMetaTexture; stdcall; external '4dxdll.dll';
function  dx_gettexture(texture : integer) : pointer; stdcall; external '4dxdll.dll';
function  dx_gettexturewidth (texture : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_gettextureheight(texture : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_getflat(flat : integer) : pointer; stdcall; external '4dxdll.dll';

function  dx_getbitmap(bitmapno : integer) : pointer; stdcall; external '4dxdll.dll';
function  dx_getbitmapwidth(bitmapno : integer) : integer; stdcall; external '4dxdll.dll';
function  dx_getbitmapheight(bitmapno : integer) : integer; stdcall; external '4dxdll.dll';

function  dx_getflatwidth(flat : integer) : integer; stdcall; external '4dxdll.dll';

procedure dx_initlevel; stdcall; external '4dxdll.dll';
function  dx_loadlevel(fname : pchar) : integer; stdcall; external '4dxdll.dll';
function  dx_savelevel(fname : pchar) : integer; stdcall; external '4dxdll.dll';
function  dx_importwad(fname,levelname : pchar) : integer; stdcall; external '4dxdll.dll';
function  dx_povactor(actor:integer)  : integer; stdcall; external '4dxdll.dll';
function  dx_getangle(x,y:single)     : integer; stdcall; external '4dxdll.dll';
function  dx_fixangle(angle : integer): integer; stdcall; external '4dxdll.dll';
function  dx_movex(x:single; angle : integer; distance: single)     :single; stdcall; external '4dxdll.dll';
function  dx_movey(y:single; angle : integer; distance: single)     :single; stdcall; external '4dxdll.dll';

//
function dx_resize(w,h:integer;databuff:pointer;pitch:integer): integer; stdcall; external '4dxdll.dll';
function dx_cleanup : integer; stdcall; external '4dxdll.dll';

//
procedure dx_positionactor(actor : integer; x,y,z : single; angle : integer); stdcall; external '4dxdll.dll';
function  dx_teleportactor  : integer; stdcall; external '4dxdll.dll';
function  dx_moveactor(actor: integer; step : single) : integer; stdcall; external '4dxdll.dll';
procedure dx_spinactor(actor,step : integer); stdcall; external '4dxdll.dll';
procedure dx_checkactor(actor:integer);  stdcall; external '4dxdll.dll';
function  dx_findsector(x,y : single) : integer;   stdcall; external '4dxdll.dll';
procedure dx_flushtexturecache; stdcall; external '4dxdll.dll';

// runtime routines for map mods.
function dx_raisesectorfloor(sector,height,speed : integer)  : integer; stdcall; external '4dxdll.dll';
function dx_raisesectorceiling(sector,height,speed :integer) : integer; stdcall; external '4dxdll.dll';
function dx_spawnmissile                                     : integer; stdcall; external '4dxdll.dll';

//functions for managing palettes
function  dx_assignpalette(palette : pointer)  : integer; stdcall; external '4dxdll.dll';
procedure dx_normalshade(r,g,b,distance : integer);       stdcall; external '4dxdll.dll';
procedure dx_alternateshade(r,g,b,distance : integer);    stdcall; external '4dxdll.dll';

// Map editing routines
function  dx_addsector(sector,floortext,ceiltext : integer; floorht, ceilht : single) : pSector; stdcall; external '4dxdll.dll';
function  dx_addactor(num,actortype: integer;x,y,ht : single; sector : integer)  : pActor; stdcall; external '4dxdll.dll';
function  dx_addline(sector,fromvert,tovert,texture : integer)   : pLineDef; stdcall; external '4dxdll.dll';
function  dx_addvertex(num:integer;x,y : single) : pVertex; stdcall; external '4dxdll.dll'; //assigns to num. returns pointer.
function  dx_insertvertex(x,y : single) : integer; stdcall; external '4dxdll.dll'; // returns where it put it

procedure dx_fixupsector(sectorno : integer); stdcall; external '4dxdll.dll';
procedure dx_fixuplines;  stdcall; external '4dxdll.dll';
procedure dx_fixupactors; stdcall; external '4dxdll.dll';
procedure dx_fixline(l : pLineDef); stdcall; external '4dxdll.dll';
procedure dx_fixactor(a : pActor);  stdcall; external '4dxdll.dll';
function  dx_getfacingline(actno : integer; distance : single) : pLineDef; stdcall external '4dxdll.dll';

type lvect = function(pa : pactor; pl : plinedef) : integer; stdcall;
type avect = function(pa,hitactor : pactor) : integer; stdcall;
type svect = function(pa : pactor; ps : psector) : integer; stdcall;
type mvect = function(pa : pactor; x,y : integer) : integer; stdcall;

// Information functions, Returns pointers to 4dx entities.
function dx_findactor(x,y:integer) : integer; stdcall; external '4dxdll.dll';
function dx_shownsectors : pointer;           stdcall; external '4dxdll.dll';
function dx_shownactorcount : integer;        stdcall; external '4dxdll.dll';
function dx_shownactors : pointer;            stdcall; external '4dxdll.dll';
function  dx_getactor(actnum : integer)  : pactor;  stdcall; external '4dxdll.dll';
function  dx_getsector(secnum : integer) : pSector; stdcall; external '4dxdll.dll';
function  dx_getvertex(vnum: integer)    : pVertex; stdcall; external '4dxdll.dll';
procedure dx_setlinecallback(lv : lvect); stdcall; external '4dxdll.dll';
procedure dx_setactorcallback(av : avect); stdcall; external '4dxdll.dll';
procedure dx_setsectcallback(sv : svect); stdcall; external '4dxdll.dll';
procedure dx_setmissilecallback(mv : mvect); stdcall; external '4dxdll.dll';

// functions created in this file:
procedure dx_LoadTextureDefs(filename : string);
procedure dx_LoadBitmapSet  (filename : string);
procedure dx_LoadFlatSet    (filename : string);
procedure dx_SaveBitmapset  (fname    : string);
procedure dx_SaveTextureDefs(fname    : string);

function  WadGetLevelnames(wadname : string) : integer;


var levelnames : array[1..40] of string;

implementation

uses sysutils,windows;


function PointerAdd(inptr : pointer; value : integer) : pointer;
var i : integer;
    P : Pointer;
begin
    P := inptr;
    i := value;
    asm
       mov ecx,p
       add ecx,value
       mov p,ecx
    end;
    result := P;
end;

function WadGetLevelnames(wadname : string) : integer;
var f: file;
    HEADER: Wad_Header;
    DIRECTORY: Directory_Entry;
    i,numlevs : integer;
    buffer : array [0..256] of char;
    EntriesRead : integer;
    EntryOffset : integer;
begin
    for i := 1 to 40 do levelnames[i] := '';
    EntriesRead := 0;
    numlevs := 0;
    assignfile(f,wadname);
    reset(f,1);
    BlockRead(F, header, 12);
    EntryOffset := header.foffset;
    while ( EntriesRead < HEADER.num_entries ) do begin
      seek(F, EntryOffset);
      BlockRead(F, DIRECTORY, sizeof(Directory_Entry));
      inc(EntriesRead);
      inc(EntryOffset,sizeof( Directory_Entry ));
      if ((DIRECTORY.name[0] = 'E') and  (DIRECTORY.name[2] = 'M')) then begin
         strlcopy(buffer,pchar(@DIRECTORY.name),4);
         buffer[4] := #0;
         levelnames[numlevs+1] := buffer;
         inc (numlevs);
      end;
      if ((DIRECTORY.name[0] = 'M') and (DIRECTORY.name[1] = 'A') and (DIRECTORY.name[2] = 'P')) then begin
         strlcopy(buffer,pchar(@DIRECTORY.name),5);
         buffer[5] := #0;
         levelnames[numlevs+1] := buffer;
         inc (numlevs);
      end;
   end;
   closefile(f);
   result := numlevs;
end;

procedure dx_LoadFlatSet(filename : string);
var i,flatnum : integer;
    f : file;
    buf : array [0..255] of char;
    flatsize : integer;
    flatdata : pByte;
    flatname : string;
begin
    AssignFile(f,filename);
    reset(f,1);
    try
       while not eof(f) do begin
          BlockRead(F, flatnum, 4);
          BlockRead(f, flatsize,4);
          BlockRead(f, Buf, 32);
          flatname := buf;
          BlockRead(f, Buf, 256);
          GetMem(flatdata,flatsize*flatsize);
          BlockRead(f, FlatData^,flatsize*flatsize);
          dx_createflat(flatnum,flatsize,flatsize,flatdata,pchar(flatname));
          FreeMem(Flatdata);
       end;
    finally
       CloseFile(f);
    end;
end;

procedure SaveFlatset(fname : string);
var i,j,flatnum : integer;
    f : file;
    buf : array [0..255] of Char;
    flatsize : integer;
    flatdata : pbyte;
    tempdata : pbyte;
    dest,src : pointer;
begin
    // save a flat set.
    AssignFile(f,fname);
    rewrite(f,1);
    try
       for i := 1 to dx_maxflats do begin
           flatnum  := i-1;
           flatsize := dx_getflatwidth(flatnum);
           if flatsize > 0 then begin
              flatdata := dx_getflat(flatnum);
              BlockWrite(f,flatnum,4);
              BlockWrite(f,flatsize,4);

              ZeroMemory(@(buf[0]),33);
              StrLCopy(buf,dx_getflatname(flatnum),8);
              BlockWrite(F, Buf, 32);

              ZeroMemory(@(buf[0]),256);
              BlockWrite(F, Buf, 256);
              GetMem(tempdata,flatsize*flatsize);
              src := flatdata;
              dest := tempdata;
              for j := 0 to flatsize-1 do begin
                  CopyMemory(dest,src,flatsize);
                  src := pointeradd(src,256);
                  dest := pointeradd(dest,flatsize);
              end;
              BlockWrite(f, TempData^,flatsize*flatsize);
              FreeMem(tempdata);
          end;
       end;
    finally
       CloseFile(f);
    end;
end;

procedure dx_SaveTextureDefs(fname : string);
var f : file;
    i,j : integer;
    pm : pmetatexture;
    smt : smalltext;
    pp  : pPatch;
    smp : smallpatch;
begin
    AssignFile(f,fname);
    rewrite(f,1);
    try
    for i := 0 to dx_maxtextures-1 do begin
        // write out each texture.
        // write out each patch for each texture
        pm := dx_getmetatexture(i);
        if (pm = nil) or (integer(pm) = -1) then begin
           smt.number := i;
           smt.w := 0;
           smt.h := 0;
           smt.numpatches := 0;
        end else begin
           smt.number := i;
           smt.w      := pm.w;
           smt.h      := pm.h;
           smt.numpatches := pm.numpatches;
        end;
        ZeroMemory(@(smt.name[0]),8);
        strlcopy(smt.name,dx_gettexturename(i),8);
        BlockWrite(f,smt,24);
        if smt.numpatches > 0 then begin
           for j := 1 to smt.numpatches do begin
               pp := pm.patcharray[j];
               smp.x    := pp.x;
               smp.y    := pp.y;
               smp.bmno := pp.bitmapno;
               smp.txno := i;
               BlockWrite(f,smp,16);
           end;
        end;
    end;
    finally
    CloseFile(f);
    end;
end;

procedure dx_SaveBitmapset(fname : string);
var i,Bitmapnum : integer;
    f : file;
    buf : array [0..255] of Char;
    x,y,w,h,Bitmapwidth,bitmapheight : integer;
    Bitmapdata : pbyte;
    tempdata : pbyte;
    dest,src : pbyte;
    pc : pchar;
begin
    // save a Bitmap set.
    AssignFile(f,fname);
    rewrite(f,1);
    try
       for i := 1 to dx_maxBitmaps do begin
           // write Bitmap Number, Bitmap name,
           // Bitmap location, Bitmap size, Bitmap data
           Bitmapnum    := i-1;
           Bitmapwidth  := dx_getBitmapwidth (Bitmapnum);
           if bitmapwidth < 1 then continue;
           BitmapHeight := dx_getBitmapHeight(Bitmapnum);
           Bitmapdata   := dx_getBitmap      (Bitmapnum);
           pc           := dx_getbitmapname  (BitmapNum);
           w := BitmapWidth;
           h := BitmapHeight;
           BlockWrite(f,Bitmapnum,4);
           BlockWrite(f,Bitmapwidth,4);
           BlockWrite(f,Bitmapheight,4);

           ZeroMemory(@(buf[0]),33);
           CopyMemory(@(buf[0]),pc,32);
           BlockWrite(F, Buf, 32);

           ZeroMemory(@(buf[0]),256);
           BlockWrite(F, Buf, 256);

           GetMem(tempdata,Bitmapwidth*Bitmapheight);
           src := bitmapdata;
           dest := tempdata;
           for x := 0 to w-1 do begin
              src := pointeradd(bitmapdata,x*h);
              for y := 0 to h-1 do begin
                  dest := PointerAdd(tempdata,(y*w)+x);
                  dest^ := src^;
                  src := pointeradd(src,1);
              end;
           end;
           BlockWrite(f, TempData^,Bitmapwidth*Bitmapheight);
           FreeMem(tempdata);
       end;
    finally
       CloseFile(f);
    end;
end;

procedure dx_LoadBitmapSet(filename : string);
var i,Bitmapnum : integer;
    f : file;
    buf : array [0..255] of char;
    BitmapWidth, BitmapHeight : integer;
    Bitmapdata : pByte;
    BitmapName : string;
begin
    AssignFile(f,filename);
    reset(f,1);
    try
       while not eof(f) do begin
          BlockRead(F, Bitmapnum, 4);
          BlockRead(f, BitmapWidth,4);
          BlockRead(f, BitmapHeight,4);
          BlockRead(f, Buf, 32);
          BitmapName := Buf;
          BlockRead(f, Buf, 256);
          GetMem(Bitmapdata,BitmapWidth*BitmapHeight);
          BlockRead(f, BitmapData^,BitmapHeight*BitmapWidth);
          dx_addbitmap(bitmapnum,bitmapwidth,bitmapheight,bitmapwidth,1,bitmapdata,pchar(BitmapName));
          FreeMem(Bitmapdata);
       end;
    finally
       CloseFile(f);
    end;
end;

procedure dx_LoadTextureDefs(filename : string);
var f : file;
    i,j : integer;
    pm : pmetatexture;
    smt : smalltext;
    pp  : pPatch;
    smp : smallpatch;
    buf : array [0..8] of char;
begin
    AssignFile(f,filename);
    reset(f,1);
    try
       for i := 0 to dx_maxtextures-1 do begin
           BlockRead(f,smt,24);
           if eof(f) then break;
           Zeromemory(@(buf[0]),9);
           strlcopy(buf,smt.name,8);
           dx_createtexture(smt.number,smt.w,smt.h,buf);
           if smt.numpatches > 0 then begin
               for j := 1 to smt.numpatches do begin
                   BlockRead(f,smp,16);
                   dx_addpatch (smp.txno,smp.bmno,smp.x,smp.y);
               end;
           end;
       end;
    finally
       CloseFile(f);
    end;
end;


end.
