unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  RDX, StdCtrls, ExtCtrls, mmsystem, dxint;

type
  TForm1 = class(TForm)
    RenderDX1: TRenderDX;
    Memo1: TMemo;
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    original_pal : hPalette;
    procedure AppIdle(Sender: TObject; var Done: Boolean);
    procedure MoveStuff;
    procedure MoveRobots;
  public
    // flags for direction of movement;
    Rotating  : integer;
    Moving    : integer;
    Shearing  : integer;

    tickcount : integer;         // this is the frame tick count.
    tickerr   : integer;         // we accumulate errors here, so that the
                                 // movement is smooth.
    StartTime,EndTime : integer; // For Timing
    robomove  : integer;          // timing robot moves.

    // Amounts of movement
    Shear     : integer;   // looking up and down
    Velocity  : integer;   // moving forward/backward
    Spin      : integer;   // angles turning
    MaxVelocity  : integer;

    SectorList : TList;
    function GetPalette : hPalette; override;
  end;

type pMoveList = ^movelist;
     movelist = record
        raisingc  : boolean;
        raisingf  : boolean;
        loweringc : boolean;
        loweringf : boolean;
        ps        : psector;
        left      : integer;
     end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

function TForm1.GetPalette : hPalette;
begin
   result :=  hPalette(renderdx1.palette);
end;

procedure TForm1.MoveRobots;
var i : integer;
    us,pa : pactor;
    dist : single;
    dx,dy : single;
    movem : boolean;
begin
    movem := false;
    us := dx_getactor(0);

    robomove := robomove + tickcount;
    if robomove > 8 then begin
        dec(robomove,8);
        movem := true;
    end;

    // Skip actor 1 since it's the pov character.
    for i := 1 to dx_maxactors-1 do begin
       pa := dx_getactor(i);
       if pa.used <> 0 then begin
          if (pa.actortype = 1) and movem then begin
             // determine the distance we are from the robot.
             dx := us.x - pa.x;
             dy := us.y - pa.y;
             dist := sqrt((dx*dx)+(dy*dy));
             if (dist < 768) and (dist > 128) then begin
                pa.adjust := -16;
                pa.angle  := dx_getangle(dx,dy);
                inc(pa.currframe);
                if pa.currframe > 4 then pa.currframe := 0;
                RenderDX1.MoveActor(i,16);
             end else begin
                pa.currframe := 3;
             end;
          end;
          dx_checkactor(i);
       end;
    end;
end;

procedure TForm1.MoveStuff;
var NeedPack : boolean;
    pm : pMoveList;
    i,amt : integer;
begin
   if sectorlist.count > 0 then begin
      NeedPack := false;
      // lets check all the sectors in this list and see if they need
      // to be raised or lowered.
      for i := 0 to sectorlist.count-1 do begin
          pm := sectorlist.items[i];
          amt := tickcount;
          if amt > pm.left then amt := pm.left;

          if pm.raisingf then begin
             pm.ps.floorht := pm.ps.floorht + amt;
          end;

          if pm.raisingc then begin
             pm.ps.ceilht := pm.ps.ceilht + amt;
          end;

          if pm.loweringf then begin
             pm.ps.floorht := pm.ps.floorht - amt;
          end;

          if pm.loweringc then begin
             pm.ps.ceilht := pm.ps.ceilht - amt;
          end;

          dec(pm.left,amt);
          if pm.left < 0 then begin
             // free pm
             Dispose(pm);
             sectorlist.delete(i);
             Needpack := true;
          end;
      end;
      // if any were removed from the list, then the list is packed.
      if NeedPack then SectorList.Pack;
   end;
end;

procedure TForm1.AppIdle(Sender: TObject; var Done: Boolean);
var ctr : integer;
    remain : integer;
    erraccum : integer;
begin
    Done := False;
    StartTime := EndTime;
    RenderDX1.Render(0,0);
    Application.ProcessMessages;
    RenderDX1.Repaint;
    EndTime   := TimeGetTime; // count in milliseconds
    ctr       := EndTime - StartTime; // ctr = # of milliseconds since last render
    TickCount := ctr shr 4;           // div 16
    remain    := ctr - (ctr  shl 4);  // mod 16
    tickerr   := Tickerr + remain;    // accumulate error
    if tickerr >= 16 then begin       // pull 'whole ticks' off err
       erraccum  := tickerr shl 4;
       tickcount := tickcount + erraccum;
       tickerr   := tickerr - (erraccum shr 4);
    end;
    // First take care of mundane tasks.
    // Move any doors / elevators that need to move.
    MoveStuff;
    // move any actors that need to chase me.
    MoveRobots;

    if moving <> 0 then begin
       // speed up
       inc(velocity,moving*tickcount);
       if velocity >  maxvelocity then velocity :=  maxvelocity;
       if velocity < -maxvelocity then velocity := -maxvelocity;
    end else begin
       // slow down
       if velocity > 0 then begin
          if velocity >= tickcount then dec(velocity,tickcount) else velocity := 0;
       end;
       if velocity < 0 then begin
          if velocity <= (-tickcount) then inc(velocity,tickcount) else velocity := 0;
       end;
    end;

    if rotating <> 0 then begin
       // speed up rotation
       inc(spin,rotating*tickcount);
       if spin >  12 then spin :=  12;
       if spin < -12 then spin := -12;
    end else begin
       // slow down rotation
       if spin >  0 then begin
          if spin >= tickcount then dec(spin,tickcount) else spin := 0;
       end;
       if spin <  0 then begin
         if spin <= (-tickcount)then inc(spin,tickcount) else spin := 0;
       end;
    end;

    if shearing <> 0 then begin
       // Look farther up and down
       inc(shear,shearing*tickcount*4);
       if shear >  200 then shear :=  200;
       if shear < -200 then shear := -200;
    end else begin
       // Move view back toward center.
       if shear >  0 then
          if shear >= (tickcount *4) then dec(shear,tickcount*4) else shear := 0;
       if shear <  0 then
          if shear <= (tickcount *-4) then inc(shear,tickcount*4) else shear := 0;
    end;
    // move viewpoint, spin, and shear the view.
    if velocity <> 0 then begin
       RenderDX1.MoveActor(0,velocity);
    end;
    if spin <> 0 then begin
       RenderDX1.SpinActor(0,Spin);
    end;
    RenderDX1.Shear := shear;
    RenderDX1.CheckActor(0);
end;

// DoType = 1000 = raise ceiling of sector
// saved height in tripflags1
// DoTag  = 111 = use sector(s) with tag 111

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var pa : pactor;
    pl : pLineDef;
    pm : pMoveList;
    ps : pSector;
    i  : integer;
begin
     case Key of
         VK_SPACE: begin
                   // Jump!
                   pa := dx_GetActor(0);
                   pa.vertforce := -20;
                   end;
        VK_RETURN: begin
                   // Activate something.
                   pl := dx_getfacingline(0,128);
                   // see if we are actually looking at a line
                   if pl <> nil then begin
                       // see if it has a "trip activate" flag on it.
                      if (pl.flags and 4) = 4 then begin
                          // remove the "trip activate" if it is a "trip once" line.
                          if (pl.tripflags and $10) = $10 then pl.flags := pl.flags and (not 4);
                          // I've only set up one tag type - 1000.
                          if pl.dotype = 1000 then begin
                             // Tag type 1000 will raise a sectors ceiling by what is in trip1.
                             for i := 0 to dx_maxsectors -1 do begin
                                ps := dx_getsector(i);
                                if ps <> nil then
                                if ps.tag = pl.dotag then begin
                                   New(PM);
                                   pm.ps := ps;
                                   pm.left := pl.trip1;
                                   pm.raisingc := true;
                                   pm.raisingf := false;
                                   pm.loweringc := false;
                                   pm.loweringf := false;
                                   SectorList.Add(pm);
                                end;
                             end;
                          end;
                      end;
                   end;
                   end;
         VK_UP:    moving   :=  1;
         VK_DOWN:  moving   := -1;
         VK_LEFT:  rotating := -1;
         VK_RIGHT: rotating :=  1;
         VK_PRIOR: shearing := -1;
         VK_NEXT:  shearing :=  1;
     end;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     case Key of
         VK_UP:    moving   :=  0;
         VK_DOWN:  moving   :=  0;
         VK_LEFT:  rotating :=  0;
         VK_RIGHT: rotating :=  0;
         VK_PRIOR: shearing :=  0;
         VK_NEXT:  shearing :=  0;
     end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var dirname : string;
begin
    Sectorlist := TList.Create;
    dirname := ExtractFilePath(application.exename);
    dirname := dirname+'..\demogrx\';
    MaxVelocity := 24;
    RenderDX1.FlatSet    := dirname+'default.flt';   // or whatever you saved the flatset as.
    RenderDX1.BitmapSet  := dirname+'default.bst';   // bitmap set, whatever you saved it as.
    RenderDX1.TextureSet := dirname+'default.tdf';   // texture definitions
    RenderDX1.ActorDefs  := dirname+'default.typ';   // Actor Type definitions
    RenderDX1.Level      := dirname+'demo2.4dx'; // the actual map level
    EndTime := TimeGetTime;
    Application.OnIdle := AppIdle;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
    Application.OnIdle := nil;
    SectorList.Free;
end;

end.
